r"""
This files generates the results for Table (4) in the BFit paper.

Specifically, it optimizes the least-squares (spherically-averaged) of the
 atomic density.  The un-normalized Gaussian model is used and the choice
 of adding the constraint that the integral of the model density should
 be equal to the integral of the atomic density is used with the variable
 `WITH_CONSTRAINT`. The universal Gaussian basis set exponents multipled by
 two are used as the initial guess.

"""
import numpy as np

from bfit.density import SlaterAtoms
from bfit.fit import ScipyFit
from bfit.grid import ClenshawRadialGrid
from bfit.measure import SquaredDifference
from bfit.model import AtomicGaussianDensity
from bfit.parse_ugbs import get_ugbs_exponents

WITH_CONSTRAINT = False

for atomic_numb, element in [
    (2, "he"),
    (4, "be"),
    (9, "f"),
    (13, "al"),
    (19, "k"),
    (22, "ti"),
    (30, "zn"),
    (45, "rh"),
    (53, "i")
][8:]:
    print("Element %s " % element)

    # Define Integration Grid
    grid = ClenshawRadialGrid(
        atomic_numb, num_core_pts=10000, num_diffuse_pts=899, extra_pts=[50, 75, 100]
    )

    # Use the same number of s-type and p-type functions as UGBS
    ugbs = get_ugbs_exponents(element)
    exps_s = ugbs["S"]
    exps_p = ugbs["P"]
    num_s = len(exps_s)
    num_p = len(exps_p)

    # Define Un-normalized Gaussian Model
    model = AtomicGaussianDensity(grid.points, num_s=num_s, num_p=num_p, normalize=False)

    # Atomic Density
    dens_obj = SlaterAtoms(element)
    density = dens_obj.atomic_density(grid.points)

    # Fit object with least-squares as the objective function.
    measure = SquaredDifference()
    fit = ScipyFit(grid, density, model, measure=measure, method="SLSQP", spherical=True)

    # Define UGBS initial guess
    c0 = np.array([atomic_numb / (num_s + num_p)] * (num_s + num_p))
    e_0 = exps_s + exps_p
    e_0 = np.array(e_0)
    e_0 *= 2.0

    # Run the optimization algorithm
    results = fit.run(
        c0, e_0, maxiter=10000, disp=True, with_constraint=WITH_CONSTRAINT, tol=1e-14
    )

    print("SLSQP INFO")
    print("---------")
    print("Success %s" % results["success"])
    print("Final Coeffs")
    print(results["coeffs"])
    print("Final Exponents")
    print(results["exps"])
    print("Integration Value & L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    p = results["performance"]
    print(p)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('../'))


# -- Project information -----------------------------------------------------

project = 'BFit'
copyright = '2021, The QC-Devs Community'
author = 'The QC-Devs Community'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "numpydoc",
    # 'sphinx.ext.autosummary',
    "sphinx.ext.doctest",
    # for adding “copy to clipboard” buttons to all text/code boxes
    "sphinx_copybutton",
    "autoapi.extension",
    "nbsphinx",
    # 'sphinxcontrib.bibtex',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# Autoapi options
# can be found in https://sphinx-autoapi.readthedocs.io/en/latest/reference/config.html
autoapi_generate_api_docs = True
autoapi_type = 'python'
autoapi_dirs = ['../bfit/']
autoapi_ignore = ["*/test_*.py"]
autoapi_python_class_content = "both"
autodoc_member_order = 'bysource'
autoapi_options = [
    "members", "inherited-members", "undoc-members", "show-inheritance",
    "show-module-summary",
]


# Napolean settings to allow documentation of __init__
napoleon_include_special_with_doc = True
napoleon_include_init_with_doc = True



r"""
This files generates the results for Table (1) and (2) in the BFit paper.

Specifically, it optimizes the Kullback-Leibler Divergence using the fixed
 point iteration method. A normalized Gaussian model is used whose initial
 guess is the universal Gaussian basis-set multipled by two.  The constraint
 that the integral of the model should equal the atomic number is added, via
 the attribute `integral_dens`. The attribute `disp` displays the results
 at each iteration.
"""
import numpy as np
from atomdb import load

from bfit.fit import KLDivergenceFPI
from bfit.grid import ClenshawRadialGrid
from bfit.model import AtomicGaussianDensity
from bfit.parse_ugbs import get_ugbs_exponents

# change the path to the location of the AtomDB data as needed
DATAPATH = ("/home/ali-tehrani/SoftwareProjects/AtomDBdata",)

results_final = {}
atoms = ["H",  "C", "N", "O", "F", "P", "S", "Cl"]
atomic_numbs = [1, 6, 7, 8, 9, 15, 16, 17]  # [1 + i for i in range(0, len(atoms))]
mult = [2, 3, 4, 3, 2, 4, 3, 2]
for k, element in enumerate(atoms):
    print("Start Atom %s" % element)

    # Construct a integration grid
    atomic_numb = atomic_numbs[k]
    grid = ClenshawRadialGrid(
        atomic_numb,
        num_core_pts=10000,
        num_diffuse_pts=899,
        include_origin=True,
        # extra_pts=[50,75,100],
    )

    # Initial Guess constructed from UGBS
    ugbs = get_ugbs_exponents(element)
    exps_s = ugbs["S"]
    num_s = len(exps_s)
    exps_p = ugbs["P"]
    num_p = len(exps_p)
    coeffs = np.array([atomic_numb / (num_s + num_p)] * (num_s + num_p))
    e_0 = np.array(exps_s + exps_p) * 2.0

    # Construct Atomic Density and Fitting Object
    # density = SlaterAtoms(element=element).atomic_density(grid.points)
    # Use AtomDB to calculate the density
    atom = load(
        elem=element,
        charge=0,
        mult=mult[k],
        dataset="hci",
        datapath=DATAPATH,
    )

    dens = atom.dens_func()
    density = dens(grid.points)

    # plt.plot(grid.points, density, "bo-")
    # plt.show()
    print(density[-1], grid.points[-1], grid.points[0:3])
    density[density < 0.0] = 0.0
    # continue
    # assert 1 == 0
    model = AtomicGaussianDensity(grid.points, num_s=num_s, num_p=num_p, normalize=True)
    fit = KLDivergenceFPI(grid, density, model, mask_value=1e-18, spherical=True,
                          integral_dens=atomic_numb)

    # Run the Kullback-Leibler FPI Method
    results = fit.run(
        coeffs, e_0, maxiter=10000, c_threshold=1e-6, e_threshold=1e-6, d_threshold=1e-14,
        disp=True
    )
    # Construct Fitting Object using SLSQP and optimizing KL
    # measure = KLDivergence(mask_value=1e-18)
    # fit_KL_slsqp = ScipyFit(grid, density, model, measure=measure, method="SLSQP", spherical=True)
    # # Run the SLSQP optimization algorithm
    # results = fit_KL_slsqp.run(
    # coeffs, e_0, maxiter=10000, disp=True, with_constraint=True, tol=1e-14
    # )

    print("KL-FPI INFO")
    print("-----------")
    print("Success %s" % results["success"])
    print("Final Coeffs")
    print(results["coeffs"])
    print("Final Exponents")
    print(results["exps"])
    print("Integration Value & L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    p = results["performance"][-1]
    print(p)
    # Calculate the relative errors
    spherical = 4.0 * np.pi * grid.points**2.0
    l1 = results["performance"][-1][1] / grid.integrate(density * spherical)
    linf = results["performance"][-1][2] / np.max(density)
    ls = results["performance"][-1][3] / grid.integrate(density**2.0 * spherical)
    kl = results["performance"][-1][4] / grid.integrate(
        density * np.log(density) * spherical
    )
    print("Relative Errors L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    print([l1, linf, ls, kl])

    # Store the results
    results_final[element + "_coeffs_s"] = results["coeffs"][:num_s]
    results_final[element + "_exps_s"] = results["exps"][:num_s]
    results_final[element + "_coeffs_p"] = results["coeffs"][num_s:]
    results_final[element + "_exps_p"] = results["exps"][num_s:]
    results_final[element + "_errors_4pir2"] = results["performance"][-1]
    results_final[element + "_sucess"] = results["success"]
    results_final[element + "_errors"] = [l1, linf, ls, kl]

    np.savez("./result_kl_fpi_method_cugbasis_atomdb_hci.npz", **results_final)



r"""
This files generates the results for Table (3) in the BFit paper.

Specifically, it optimizes the spherically-averaged Kullback-Leibler divergence
 function using the SLSQP function (from class ScipyFit) of the atomic density.
 The initial guess is chosen to be universal Gaussian basis-set (UGBS) multipled
 by two. Normalized Gaussian model is used.
"""
import numpy as np

from bfit.density import SlaterAtoms
from bfit.fit import ScipyFit
from bfit.grid import ClenshawRadialGrid
from bfit.measure import KLDivergence
from bfit.model import AtomicGaussianDensity
from bfit.parse_ugbs import get_ugbs_exponents

for atomic_numb, element in [
    (2, "he"),
    (4, "be"),
    (9, "f"),
    (13, "al"),
    (19, "k"),
    (22, "ti"),
    (30, "zn"),
    (45, "rh"),
    (53, "i")
][-1:]:
    print("\n Start Atom %s" % element)

    # Construct Integration Grid
    grid = ClenshawRadialGrid(
        atomic_numb, num_core_pts=10000, num_diffuse_pts=899, extra_pts=[50, 75, 100]
    )

    # Construct Initial Guess with UGBS
    ugbs = get_ugbs_exponents(element)
    exps_s = ugbs["S"]
    num_s = len(exps_s)
    exps_p = ugbs["P"]
    num_p = len(exps_p)
    c0 = np.array([atomic_numb / (num_s + num_p)] * (num_s + num_p))
    e0 = np.array(exps_s + exps_p) * 2.0

    model = AtomicGaussianDensity(grid.points, num_s=num_s, num_p=num_p, normalize=True)
    atomic_dens = SlaterAtoms(element=element)
    density = atomic_dens.atomic_density(grid.points)

    # Construct Fitting Object using SLSQP and optimizing KL
    measure = KLDivergence(mask_value=1e-18)
    fit_KL_slsqp = ScipyFit(grid, density, model, measure=measure, method="SLSQP", spherical=True)

    # Run the SLSQP optimization algorithm
    results = fit_KL_slsqp.run(c0, e0, maxiter=10000, disp=True, with_constraint=True, tol=1e-14)

    print("SLSQP INFO")
    print("----------")
    print("Success %s" % results["success"])
    print("Final Coeffs")
    print(results["coeffs"])
    print("Final Exponents")
    print(results["exps"])
    print("Integration Value & L1 & L_infinity & LS & KL (With 4 pi r^2 included)")
    p = results["performance"]
    print(p)



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test file for 'bfit.greedy'."""

import numpy as np
import numpy.testing as npt

from bfit.greedy import (
    get_next_choices,
    get_two_next_choices,
    GreedyKLFPI,
    GreedyLeastSquares,
    pick_two_lose_one,
    remove_redundancies,
)
from bfit.grid import UniformRadialGrid


def test_check_redundancies():
    r"""Testing for checking redundancies for greedy."""
    c = np.array([50., 30., 10., 2., 3.])
    exps = np.array([3.1, 3., 2, 1, 3.025])
    simi = 1.
    true_answer = remove_redundancies(c, exps, simi)
    desired_answer_coeff = np.array([83., 10., 2.])
    desired_answer_exps = np.array([3.1, 2., 1.])
    npt.assert_allclose(true_answer[0], desired_answer_coeff)
    npt.assert_allclose(true_answer[1], desired_answer_exps)

    c = np.array([50., 30., 10., 2., 3.])
    exps = np.array([3.0012, 3.0005, 2, 1, 3.025])
    simi = 1e-2
    true_answer = remove_redundancies(c, exps, simi)
    npt.assert_array_equal(true_answer[0], np.array([80, 10., 2., 3]))
    npt.assert_array_equal(true_answer[1], np.array([3.0012, 2., 1., 3.025]))

    simi = 1e-1
    true_answer = remove_redundancies(c, exps, simi)
    npt.assert_array_equal(true_answer[0], np.array([83, 10., 2]))
    npt.assert_array_equal(true_answer[1], np.array([3.0012, 2, 1]))


def test_get_next_possible_coeffs_and_exps():
    r"""Testing 'bfit.greedy.greedy_utils.get_next_choices'."""
    c = np.array([1., 2.])
    e = np.array([3., 4.])
    factor = 2
    true_answer = get_next_choices(factor, c, e, coeff_val=0)
    desired_answer = [[0., 1, 2, 1.5, 3, 4],
                      [1, 0, 2, 3, 3.5, 4],
                      [1, 2, 0, 3, 4, 8]]
    npt.assert_array_equal(true_answer, desired_answer)


def test_get_two_next_possible_coeffs_and_exps():
    r"""Testing 'bfit.greedy.greedy_utils.get_two_next_choices'."""
    # One Basis Function
    c = np.array([1.])
    e = np.array([3.])
    factor = 2
    true_answer = get_two_next_choices(factor, c, e, coeff_val=0)
    desired_answer = [[0, 0, 1, 0.75, 1.5, 3],
                      [0, 0, 1, 1.5, 2.25, 3],
                      [0, 1, 0, 1.5, 3, 6],
                      [1, 0, 0, 3, 4.5, 6],
                      [1, 0, 0, 3, 6, 12]]
    npt.assert_array_equal(true_answer, desired_answer)

    # Two Basis Function
    c = np.array([1., 2.])
    e = np.array([3., 4.])
    factor = 2
    true_answer = get_two_next_choices(factor, c, e, coeff_val=0)
    desired_answer = [[0, 0, 1, 2, 0.75, 1.5, 3., 4.],
                      [0, 0, 1, 2, 1.5, 2.25, 3, 4],
                      [0, 1., 0, 2, 1.5, 3, 3.5, 4.],
                      [0, 1, 2, 0, 1.5, 3, 4., 8.],
                      [1, 0, 0, 2, 3, 3.25, 3.5, 4],
                      [1, 0, 0, 2, 3, 3.5, 3.75, 4.],
                      [1, 0, 2, 0, 3, 3.5, 4., 8.],
                      [1, 2, 0, 0, 3, 4, 6, 8],
                      [1, 2, 0, 0, 3, 4, 8, 16]]
    npt.assert_array_equal(true_answer, desired_answer)


def test_pick_two_lose_one():
    r"""Testing 'bfit.greedy.greedy_utils.pick_two_lose_one'."""
    c = np.array([1])
    e = np.array([3.])
    factor = 2
    true_answer = pick_two_lose_one(factor, c, e, coeff_val=0)
    desired_answer = [[0, 1, 1.5, 3],
                      [0, 1, 0.75, 3],
                      [0, 0, 0.75, 1.5],
                      [0, 1, 2.25, 3],
                      [0, 1, 1.5, 3],
                      [0, 0, 1.5, 2.25],
                      [1, 0, 3, 6],
                      [0, 0, 1.5, 6],
                      [0, 1, 1.5, 3],
                      [0, 0, 4.5, 6],
                      [1, 0, 3, 6],
                      [1, 0, 3, 4.5],
                      [0, 0, 6, 12],
                      [1, 0, 3, 12],
                      [1, 0, 3, 6]]
    npt.assert_array_equal(true_answer, desired_answer)


def test_greedy_kl_two_function():
    r"""Test Greedy Kullback-Leibler against two-function Gaussian combination."""
    def eval_density(points):
        return 0.25 * np.exp(-10. * points**2.0) * (10.0 / np.pi)**1.5 + \
                0.75 * np.exp(-5. * points**2.0) * (5.0 / np.pi)**1.5

    grid = UniformRadialGrid(1000, 0.0, 10.)
    density = eval_density(grid.points)
    greedy = GreedyKLFPI(grid, density, "pick-one",
                         g_eps_coeff=1e-10, g_eps_exp=1e-10,
                         l_eps_exp=1e-5, g_eps_obj=1e-15,
                         l_eps_obj=1e-10, l_eps_coeff=1e-8, mask_value=0.0,
                         l_maxiter=15000, g_maxiter=15000, integral_dens=1.0,
                         spherical=True)
    result = greedy.run(2.5, max_numb_funcs=2, disp=True)

    npt.assert_almost_equal(result["fun"], 0.0, decimal=6)
    npt.assert_almost_equal(np.sort(result["coeffs"]), [0.25, 0.75], decimal=3)
    npt.assert_almost_equal(np.sort(result["exps"]), [5.0, 10.0], decimal=3)
    assert result["success"]


def test_greedy_ls_two_function():
    r"""Test Greedy Least-Squares against two-function Gaussian combination."""
    def eval_density(points):
        return 0.25 * np.exp(-10. * points**2.0) * (10.0 / np.pi)**1.5 + \
                0.75 * np.exp(-5. * points**2.0) * (5.0 / np.pi)**1.5

    grid = UniformRadialGrid(1000, 0.0, 10.)
    density = eval_density(grid.points)
    greedy = GreedyLeastSquares(
        grid, density, "pick-one", local_tol=1e-10, global_tol=1e-15,
        integral_dens=1.0, normalize=True, spherical=True,
    )
    result = greedy.run(2.5, d_threshold=1e-10, max_numb_funcs=2, disp=True)

    npt.assert_almost_equal(result["fun"], 0.0, decimal=10)
    npt.assert_almost_equal(np.sort(result["coeffs"]), [0.25, 0.75], decimal=3)
    npt.assert_almost_equal(np.sort(result["exps"]), [5.0, 10.0], decimal=3)
    assert result["success"]



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test bfit.measure module."""

import numpy as np
from numpy.testing import assert_almost_equal, assert_raises

from bfit.measure import KLDivergence, SquaredDifference, TsallisDivergence


def test_raises_kl():
    r"""Test raise error when using Kullback-Leibler."""
    measure = KLDivergence()
    # check density argument
    assert_raises(ValueError, measure.evaluate, 3.5, np.ones((10,)))
    assert_raises(ValueError, measure.evaluate, np.array([[1., 2., 3.]]), np.ones((10,)))
    assert_raises(ValueError, measure.evaluate, np.array([[1.], [2.], [3.]]), np.ones((10,)))
    assert_raises(ValueError, measure.evaluate, np.array([-1., 2., 3.]), np.ones((10,)))
    assert_raises(ValueError, measure.evaluate, np.array([1., -2., -3.]), np.ones((10,)))
    # check model argument
    assert_raises(ValueError, measure.evaluate, np.array([1., 2., 3.]), 1.75)
    assert_raises(ValueError, measure.evaluate, np.array([1., 2., 3.]), np.array([1., 2., 3., 4.]))
    assert_raises(
        ValueError, measure.evaluate, np.array([1., 2., 3.]), np.array([[1.], [2.], [3.]])
    )
    # Assert that measure evaluate returns infinity when value is negative
    assert np.all(
        measure.evaluate(np.array([1., 2., 3.]), np.array([-0.5, -1.3, -2.8]))
        == measure.negative_val
    )


def test_evaluate_kl_equal():
    r"""Test evaluating Kullback-Leibler and its derivative against zero model."""
    # test equal density & zero model
    dens = np.array([0.0, 1.0, 2.0])
    model = np.array([0., 0., 0.])
    measure = KLDivergence(mask_value=1.e-12)
    assert_almost_equal(
        np.array([0., 0., 0.]), measure.evaluate(dens, dens, deriv=False), decimal=8
    )
    assert_almost_equal(
        np.array([0., 0., 0.]), measure.evaluate(dens, dens, deriv=True)[0], decimal=8
    )
    assert_almost_equal(
        -np.array([1., 1., 1.]), measure.evaluate(dens, dens, deriv=True)[1], decimal=8
    )
    assert_almost_equal(
        np.array([0., 0., 0.]), measure.evaluate(dens, model, deriv=False), decimal=8
    )
    assert_almost_equal(
        np.array([0., 0., 0.]), measure.evaluate(dens, model, deriv=True)[0], decimal=8
    )
    assert_almost_equal(
        -np.array([1., 1., 1.]), measure.evaluate(dens, model, deriv=True)[1], decimal=8
    )


def test_evaluate_kl():
    r"""Test evaluating Kullback-Leibler and its derivative against sympy results."""
    # test different density and model against sympy
    dens = np.linspace(0.0, 10., 15)
    model = np.linspace(0.0, 12., 15)
    # KL divergence measure
    measure = KLDivergence(mask_value=1.e-12)
    k = np.array([0., -0.1302296834, -0.2604593668, -0.3906890503, -0.5209187337, -0.6511484171,
                  -0.7813781005, -0.911607784, -1.0418374674, -1.1720671508, -1.3022968342,
                  -1.4325265177, -1.5627562011, -1.6929858845, -1.8232155679])
    dk = np.array([-1.0] + [-0.8333333333] * 14)
    assert_almost_equal(k, measure.evaluate(dens, model, deriv=False), decimal=8)
    assert_almost_equal(k, measure.evaluate(dens, model, deriv=True)[0], decimal=8)
    assert_almost_equal(dk, measure.evaluate(dens, model, deriv=True)[1], decimal=8)


def test_raises_squared_difference():
    r"""Test raises error in squared difference class."""
    ls = SquaredDifference()
    # check density argument
    assert_raises(ValueError, ls.evaluate, 3.5, np.ones((10,)))
    assert_raises(ValueError, ls.evaluate, np.array([[1., 2., 3.]]), np.ones((10,)))
    assert_raises(ValueError, ls.evaluate, np.array([[1.], [2.], [3.]]), np.ones((10,)))
    # check model argument
    assert_raises(ValueError, ls.evaluate, np.array([1., 2., 3.]), 1.75)
    assert_raises(ValueError, ls.evaluate, np.array([1., 2., 3.]), np.array([1., 2., 3., 4.]))
    assert_raises(ValueError, ls.evaluate, np.array([1., 2., 3.]), np.array([[1.], [2.], [3.]]))


def test_evaluate_squared_difference_equal():
    r"""Test evaluating square difference on a zero model."""
    # test equal density & zero model
    dens = np.array([0.0, 1.0, 2.0])
    model = np.array([0., 0., 0.])
    measure = SquaredDifference()
    assert_almost_equal(np.zeros(3), measure.evaluate(dens, dens, deriv=False), decimal=8)
    assert_almost_equal(np.zeros(3), measure.evaluate(dens, dens, deriv=True)[0], decimal=8)
    assert_almost_equal(np.zeros(3), measure.evaluate(dens, dens, deriv=True)[1], decimal=8)
    assert_almost_equal(dens**2, measure.evaluate(dens, model, deriv=False), decimal=8)
    assert_almost_equal(dens**2, measure.evaluate(dens, model, deriv=True)[0], decimal=8)
    assert_almost_equal(-2 * dens, measure.evaluate(dens, model, deriv=True)[1], decimal=8)


def test_evaluate_squared_difference():
    r"""Test evaluating the squared difference and its derivative against sympy."""
    # test different density and model against sympy
    dens = np.linspace(0.0, 10., 15)
    model = np.linspace(0.0, 12., 15)
    # KL divergence measure
    measure = SquaredDifference()
    # equal density
    assert_almost_equal(np.zeros(15), measure.evaluate(dens, dens, deriv=False), decimal=8)
    assert_almost_equal(np.zeros(15), measure.evaluate(dens, dens, deriv=True)[0], decimal=8)
    assert_almost_equal(np.zeros(15), measure.evaluate(dens, dens, deriv=True)[1], decimal=8)
    # different densities
    m = np.array([0.0, 0.0204081633, 0.0816326531, 0.1836734694, 0.3265306122,
                  0.5102040816, 0.7346938776, 1.0, 1.306122449, 1.6530612245,
                  2.0408163265, 2.4693877551, 2.9387755102, 3.4489795918, 4.0])
    dm = np.array([0.0, 0.2857142857, 0.5714285714, 0.8571428571, 1.1428571429,
                   1.4285714286, 1.7142857143, 2.0, 2.2857142857, 2.5714285714,
                   2.8571428571, 3.1428571429, 3.4285714286, 3.7142857143, 4.0])
    assert_almost_equal(m, measure.evaluate(dens, model, deriv=False), decimal=8)
    assert_almost_equal(m, measure.evaluate(dens, model, deriv=True)[0], decimal=8)
    assert_almost_equal(dm, measure.evaluate(dens, model, deriv=True)[1], decimal=8)


def test_raises_error_tsallis():
    r"""Test raises error when using Tsallis measure."""
    measure = TsallisDivergence()
    # check density argument
    assert_raises(ValueError, measure.evaluate, 3.5, np.ones((10,)))
    assert_raises(ValueError, measure.evaluate, np.array([[1., 2., 3.]]), np.ones((10,)))
    assert_raises(ValueError, measure.evaluate, np.array([[1.], [2.], [3.]]), np.ones((10,)))
    # check model argument
    assert_raises(ValueError, measure.evaluate, np.ones((10,)), 1.75)
    assert_raises(ValueError, measure.evaluate, np.ones((10,)), np.array([[1.], [2.], [3.]]))
    # check alpha parameter
    assert_raises(ValueError, TsallisDivergence, -10.)
    assert_raises(ValueError, TsallisDivergence, 1.0)


def test_evaluate_tsallis():
    r"""Test evaluating the Tsallis measure."""
    # Test close to one so masking plays a role.
    num_pts = 10000
    dens = np.linspace(0.0, 1., num_pts)
    # KL divergence measure
    alpha = 1.001
    measure = TsallisDivergence(alpha=alpha, mask_value=1e-8)
    # Test when model is equal to density
    assert_almost_equal(np.zeros(num_pts), measure.evaluate(dens, dens, deriv=False), decimal=8)
    # Test against brute force
    model = np.random.random(num_pts)
    # First first element to be zero, so that masking plays an important role
    model[0] = 0.
    result = []
    for i in range(num_pts):
        # Set any model values to zero if it is zero.
        if model[i] < 1e-8:
            return 0.0
        result.append(dens[i] * ((dens[i] / model[i])**(alpha - 1) - 1.0) / (alpha - 1))
    assert_almost_equal(result, measure.evaluate(dens, model, deriv=False), decimal=8)


def test_evaluating_tsallis_derivative_against_finite_difference():
    r"""Test evalauting the Tsallis measure against finite-difference."""
    num_pts = 10000
    eps = 1e-8
    dens = np.linspace(0.0, 1., num_pts)
    model = np.random.random(num_pts)
    measure = TsallisDivergence(alpha=1.001, mask_value=0.0)

    # Approximate it with central finite difference with accuracy 4.
    measure_pt_minus_2 = measure.evaluate(dens, model - 2.0 * eps, deriv=False)
    measure_pt_minus_1 = measure.evaluate(dens, model - eps, deriv=False)
    _, actual_deriv = measure.evaluate(dens, model, deriv=True)
    measure_pt_plus_1 = measure.evaluate(dens, model + eps, deriv=False)
    measure_pt_plus_2 = measure.evaluate(dens, model + 2.0 * eps, deriv=False)
    desired = measure_pt_minus_2 / 12.0 - (2.0 / 3.0) * measure_pt_minus_1
    desired += (2.0 / 3.0) * measure_pt_plus_1 - measure_pt_plus_2 / 12.0
    assert_almost_equal(actual_deriv, desired / eps, decimal=1)


def test_evaluating_tsallis_derivative_against_kullback_leibler():
    r"""Test evaluating Tsallis derivative against Kullback-Leibler."""
    num_pts = 10000
    dens = np.linspace(0.0, 1., num_pts)
    alpha = 1.000000001  # Make alpha small to approximate Kullback-Leibler
    measure = TsallisDivergence(alpha=alpha, mask_value=1e-8)
    measure_kl = KLDivergence(mask_value=1e-8)

    model = np.random.random(num_pts)
    _, deriv_tsallis = measure.evaluate(dens, model, deriv=True)
    _, deriv_kl = measure_kl.evaluate(dens, model, deriv=True)
    assert_almost_equal(deriv_tsallis, deriv_kl, decimal=1)



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test bfit.model module."""

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal, assert_raises

from bfit.grid import CubicGrid, UniformRadialGrid
from bfit.model import AtomicGaussianDensity, MolecularGaussianDensity


def test_raises_gaussian_model():
    r"""Test raises error of all Gaussian model classes."""
    # check points
    assert_raises(TypeError, AtomicGaussianDensity, 10., None, 5, 5, False)
    assert_raises(TypeError, AtomicGaussianDensity, [5.], None, 5, 5, True)
    assert_raises(TypeError, AtomicGaussianDensity, (2, 1), None, 5, 5, True)
    # check num_s & num_p
    assert_raises(TypeError, AtomicGaussianDensity, np.array([0.]), None, 5, -1, False)
    assert_raises(TypeError, AtomicGaussianDensity, np.array([0.]), None, -5, 1, False)
    assert_raises(TypeError, AtomicGaussianDensity, np.array([0.]), None, 5, 1.0, False)
    assert_raises(TypeError, AtomicGaussianDensity, np.array([0.]), None, 5.1, 1, False)
    assert_raises(ValueError, AtomicGaussianDensity, np.array([0.]), None, 0, 0, False)
    assert_raises(ValueError, AtomicGaussianDensity, np.array([0.]), None, 0, 0, True)
    assert_raises(ValueError, AtomicGaussianDensity, np.array([0.]), [1.], 1, 0, True)
    assert_raises(ValueError, AtomicGaussianDensity, np.array([0.]), np.array([[1.]]), 1, 0, True)
    assert_raises(ValueError, AtomicGaussianDensity, np.array([[0.]]), np.array([1., 2.]), 1)
    # test on molecular density.
    assert_raises(ValueError, MolecularGaussianDensity, np.array([0.]), np.array([1.]), [1, 1])
    assert_raises(ValueError, MolecularGaussianDensity,
                  np.array([0.]), np.array([[1.]]), np.array([1, 1]))
    assert_raises(ValueError, MolecularGaussianDensity,
                  np.array([0.]), np.array([[1.], [2.]]), np.array([[1, 1]]))
    assert_raises(ValueError, MolecularGaussianDensity,
                  np.array([[0., 1.]]), np.array([[1.]]), np.array([[1, 1]]))


def test_gaussian_1s_origin():
    r"""Test evaluation of single s-type Gaussian at the origin."""
    # test one (un)normalized s-type gaussian on r=0. against sympy
    coeffs, expons = np.array([0.]), np.array([0.])
    # un-normalized at r=0.
    g, dg = np.array([0.]), np.array([[1., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g, dg = np.array([0.]), np.array([[0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test one (un)normalized s-type gaussian on r=0. against sympy
    coeffs, expons = np.array([3.]), np.array([2.])
    # un-normalized at r=0.
    g, dg = np.array([3.]), np.array([[1., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g, dg = np.array([1.52384726242178]), np.array([[0.507949087473928, 1.14288544681634]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1s_one_point():
    r"""Test evaluation of single s-type Gaussian at various single points."""
    # test one (un)normalized s-type gaussian on one point against sympy
    coeffs, expons = np.array([3.]), np.array([0.])
    # un-normalized at r=1.
    g, dg = np.array([3.]), np.array([[1., -3.]])
    model = AtomicGaussianDensity(np.array([1.]), num_s=1, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=1.
    g, dg = np.array([0.]), np.array([[0., 0.]])
    model = AtomicGaussianDensity(np.array([1.]), num_s=1, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test one (un)normalized s-type gaussian on one point against sympy
    coeffs, expons = np.array([3.15]), np.array([1.75])
    # un-normalized at r=-1.5
    g, dg = np.array([0.0614152227]), np.array([[0.0194968961, -0.1381842511]])
    model = AtomicGaussianDensity(np.array([-1.5]), num_s=1, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=-1.5
    g, dg = np.array([0.0255333792]), np.array([[0.0081058346, -0.0355643496]])
    model = AtomicGaussianDensity(np.array([-1.5]), num_s=1, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1s_multiple_point():
    r"""Test evaluation of single s-type Gaussian at multiple points."""
    # test one (un)normalized s-type gaussian on multiple point against sympy
    coeffs, expons = np.array([0.]), np.array([0.])
    # un-normalized r=[0., 0.5, 1.0]
    g, dg = np.zeros(3), np.array([[1., 0.], [1., 0.], [1., 0.]])
    model = AtomicGaussianDensity(np.array([0., 0.5, 1.]), num_s=1, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized r=[0., 0.5, 1.0]
    g, dg = np.zeros(3), np.zeros((3, 2))
    model = AtomicGaussianDensity(np.array([0., 0.5, 1.]), num_s=1, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test one (un)normalized s-type gaussian on multiple point against sympy
    coeffs, expons = np.array([5.25]), np.array([0.8])
    # un-normalized r=[0., 1.5, -2.2, 3.1, 10.]
    g = np.array([5.25, 0.8678191632, 0.1092876455, 0.0024060427, 0.0])
    dg = np.array([[1.0, 0.0], [0.1652988882, -1.9525931171], [0.0208166944, -0.5289522041],
                   [0.0004582938, -0.0231220701], [0.0, 0.0]])
    model = AtomicGaussianDensity(np.array([0., 1.5, -2.2, 3.1, 10.]), None, 1, 0, False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized r=[0., 1.5, 2.2, 3.1, 10.]
    g = np.array([0.6746359418, 0.1115165711, 0.0140436902, 0.0003091815, 0.0])
    dg = np.array([[0.1285020841, 1.2649423908], [0.0212412516, -0.0418187142],
                   [0.0026749886, -0.0416395415], [5.88917e-05, -0.0023915189], [0.0, 0.0]])
    model = AtomicGaussianDensity(np.array([0., 1.5, -2.2, 3.1, 10.]), None, 1, 0, True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_xs_origin():
    r"""Test evaluation of multiple s-type Gaussian at origin."""
    # test multiple (un)normalized s-type gaussian on r=0. against sympy
    coeffs, expons = np.zeros(10), np.zeros(10)
    # un-normalized at r=0.
    g, dg = np.array([0.]), np.array([[1.] * 10 + [0.] * 10])
    model = AtomicGaussianDensity(np.array([0.]), num_s=10, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g, dg = np.array([0.]), np.zeros(20)[None, :]
    model = AtomicGaussianDensity(np.array([0.]), num_s=10, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test multiple (un)normalized s-type gaussian on r=0. against sympy
    coeffs, expons = np.array([0.0, 1.0, 0.0, 3.5]), np.array([0.0, 0.0, 2.0, 0.5])
    # un-normalized at r=0.
    g, dg = np.array([4.5]), np.array([[1., 1., 1., 1., 0., 0., 0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=4, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g = np.array([0.2222277257])
    dg = np.array([[0., 0., 0.5079490874, 0.0634936359, 0., 0., 0., 0.6666831773]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=4, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_xs_one_point():
    r"""Test evaluation of multiple s-type Gaussian at single point."""
    # test multiple (un)normalized s-type gaussian on one point against sympy
    coeffs, expons = np.array([0.0, 1.0, 0.0, 3.5]), np.array([0.0, 0.0, 2.0, 0.5])
    # un-normalized at r=-1.
    g = np.array([3.12285730899422])
    dg = np.array([[1, 1, 0.1353352832, 0.606530659712633, 0., -1., 0., -2.1228573089]])
    model = AtomicGaussianDensity(np.array([-1.]), num_s=4, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=-1.
    g = np.array([0.1347879291])
    dg = np.array([[0., 0., 0.0687434336, 0.0385108368, 0., 0., 0., 0.2695758582]])
    model = AtomicGaussianDensity(np.array([-1.]), num_s=4, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # un-normalized at r=2.5
    g = np.array([1.1537792676])
    dg = np.array([[1., 1., 3.7266531720e-6, 0.0439369336, 0., -6.25, 0., -0.9611204230]])
    model = AtomicGaussianDensity(np.array([2.5]), num_s=4, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=2.5
    g = np.array([0.0097640048])
    dg = np.array([[0., 0., 1.8929500780e-6, 0.0027897156, 0., 0., 0., -0.0317330157]])
    model = AtomicGaussianDensity(np.array([2.5]), num_s=4, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_xs_multiple_point():
    r"""Test evaluation of multiple s-type Gaussian at multiple points."""
    # test multiple (un)normalized s-type gaussian on multiple points against sympy
    points = np.array([0.0, -2.67, 0.43, 5.1])
    coeffs = np.array([2.0, 0.02, 1.51])
    expons = np.array([6.1, 0.19, 7.67])
    # un-normalized
    g = np.array([3.53, 0.0051615725, 1.0323926829, 0.0001428204])
    dg = [[1.0, 1.0, 1.0, 0.0, 0.0, 0.0],
          [0.0, 0.258078623, 0.0, -0.0, -0.0367963339, -0.0],
          [0.3237155762, 0.9654789302, 0.2421536105, -0.1197100201, -0.0035703411, -0.0676090459],
          [0.0, 0.0071410175, 0.0, -0.0, -0.0037147573, -0.0]]
    dg = np.array(dg)
    model = AtomicGaussianDensity(points, num_s=3, num_p=0, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized
    g = np.array([11.1718777104, 7.67693e-05, 3.1468802527, 2.1242e-06])
    dg = [[2.7056395801, 0.0148732402, 3.8147689308, 1.3306424164, 0.0023484064, 1.126525636],
          [0.0, 0.0038384654, 0.0, -0.0, 5.87928e-05, -0.0],
          [0.8758576756, 0.0143598001, 0.9237600699, 0.1068575081, 0.0022142343, 0.0148793624],
          [0.0, 0.0001062101, 0.0, -0.0, -3.84805e-05, -0.0]]
    dg = np.array(dg)
    model = AtomicGaussianDensity(points, num_s=3, num_p=0, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1p_origin():
    r"""Test evaluation of single p-type Gaussian at origin."""
    # test one (un)normalized p-type gaussian on r=0. against sympy
    coeffs, expons = np.array([0.]), np.array([0.])
    # un-normalized at r=0.
    g, dg = np.array([0.]), np.array([[0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g, dg = np.array([0.]), np.array([[0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test one (un)normalized p-type gaussian on r=0. against sympy
    coeffs, expons = np.array([3.]), np.array([2.])
    # un-normalized at r=0.
    g, dg = np.array([0.]), np.array([[0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1p_one_point():
    r"""Test evaluation of single p-type Gaussian at single point."""
    # test one (un)normalized p-type gaussian on one point against sympy
    coeffs, expons = np.array([3.]), np.array([0.])
    # un-normalized at r=-1.
    g, dg = np.array([3.]), np.array([[1., -3.]])
    model = AtomicGaussianDensity(np.array([-1.]), num_s=0, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=-1.
    g, dg = np.array([0.]), np.array([[0., 0.]])
    model = AtomicGaussianDensity(np.array([-1.]), num_s=0, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test one (un)normalized p-type gaussian on one point against sympy
    coeffs, expons = np.array([3.15]), np.array([1.75])
    # un-normalized at r=1.5
    g, dg = np.array([0.1381842511]), np.array([[0.0438680162, -0.3109145651]])
    model = AtomicGaussianDensity(np.array([1.5]), num_s=0, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=1.5
    g, dg = np.array([0.0670251204]), np.array([[0.0212778160, -0.0550563489]])
    model = AtomicGaussianDensity(np.array([1.5]), num_s=0, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1p_multiple_point():
    r"""Test evaluation of single p-type Gaussian at multiple points."""
    # test one (un)normalized p-type gaussian on multiple point against sympy
    coeffs, expons = np.array([0.]), np.array([0.])
    # un-normalized r=[0., -0.5, 1.0]
    g, dg = np.zeros(3), np.array([[0., 0.], [0.25, 0.], [1., 0.]])
    model = AtomicGaussianDensity(np.array([0., -0.5, 1.]), num_s=0, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized r=[0., -0.5, 1.0]
    g, dg = np.zeros(3), np.zeros((3, 2))
    model = AtomicGaussianDensity(np.array([0., -0.5, 1.]), num_s=0, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test one (un)normalized p-type gaussian on multiple point against sympy
    coeffs, expons = np.array([5.25]), np.array([0.8])
    # un-normalized r=[0., 1.5, 2.2, -3.1, 10.]
    g = np.array([0.0, 1.9525931171, 0.5289522041, 0.0231220701, 0.0])
    dg = np.array([[0.0, 0.0], [0.3719224985, -4.3933345135],
                   [0.1007528008, -2.5601286677], [0.0044042038, -0.222203094], [0.0, 0.0]])
    model = AtomicGaussianDensity(np.array([0., 1.5, 2.2, -3.1, 10.]), None, 0, 1, False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized r=[0., 1.5, 2.2, -3.1, 10.]
    g = np.array([0.0, 0.1338198854, 0.0362514457, 0.0015846582, 0.0])
    dg = np.array([[0.0, 0.0], [0.025489502, 0.1170923997], [0.0069050373, -0.0621712293],
                   [0.0003018397, -0.0102765087], [0.0, 0.0]])
    model = AtomicGaussianDensity(np.array([0., 1.5, 2.2, -3.1, 10.]), None, 0, 1, True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_xp_origin():
    r"""Test evaluation of multiple p-type Gaussian at origin."""
    # test multiple (un)normalized p-type gaussian on r=0. against sympy
    coeffs, expons = np.zeros(10), np.zeros(10)
    # un-normalized at r=0.
    g, dg = np.array([0.]), np.array([[0.] * 20])
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=10, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g, dg = np.array([0.]), np.zeros(20)[None, :]
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=10, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # test multiple (un)normalized p-type gaussian on r=0. against sympy
    coeffs, expons = np.array([0.0, 1.0, 0.0, 3.5]), np.array([0.0, 0.0, 2.0, 0.5])
    # un-normalized at r=0.
    g, dg = np.array([0]), np.array([[0.] * 8])
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=4, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    model = AtomicGaussianDensity(np.array([0.]), num_s=0, num_p=4, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_xp_one_point():
    r"""Test evaluation of multiple p-type Gaussian at single point."""
    # test multiple (un)normalized p-type gaussian on one point against sympy
    coeffs, expons = np.array([0.0, 1.0, 0.0, 3.5]), np.array([0.0, 0.0, 2.0, 0.5])
    # un-normalized at r=1.
    g = np.array([3.12285730899422])
    dg = np.array([[1, 1, 0.1353352832, 0.606530659712633, 0., -1., 0., -2.1228573089]])
    model = AtomicGaussianDensity(np.array([1.]), num_s=0, num_p=4, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=1.
    g = np.array([0.0449293097])
    dg = np.array([[0., 0., 0.0916579114, 0.0128369456, 0., 0., 0., 0.1797172388]])
    model = AtomicGaussianDensity(np.array([1.]), num_s=0, num_p=4, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # un-normalized at r=-2.5
    g = np.array([7.2111204230])
    dg = np.array([[6.25, 6.25, 2.3291582325e-5, 0.2746058351, 0., -39.0625, 0., -6.0070026438]])
    model = AtomicGaussianDensity(np.array([-2.5]), num_s=0, num_p=4, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=-2.5
    g = np.array([0.0203416767])
    dg = np.array([[0., 0., 1.5774583984e-5, 0.0058119076, 0., 0., 0., -0.0254270959]])
    model = AtomicGaussianDensity(np.array([-2.5]), num_s=0, num_p=4, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_xp_multiple_point():
    r"""Test evaluation of multiple p-type Gaussian at multiple points."""
    # test multiple (un)normalized p-type gaussian on multiple points against sympy
    points = np.array([0.0, 2.67, -0.43, 5.1])
    coeffs = np.array([2.0, 0.02, 1.51])
    expons = np.array([6.1, 0.19, 7.67])
    # un-normalized
    g = np.array([0.0, 0.0367963339, 0.1908894071, 0.0037147573])
    dg = [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 1.8398166958, 0.0, 0.0, -0.2623173849, 0.0],
          [0.05985501, 0.1785170542, 0.0447742026, -0.0221343827, -0.0006601561, -0.0125009126],
          [0.0, 0.1857378662, 0.0, 0.0, -0.096620838, 0.0]]
    dg = np.array(dg)
    model = AtomicGaussianDensity(points, num_s=0, num_p=3, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized
    g = np.array([0.0, 6.93222e-05, 2.6359627773, 6.9984e-06])
    dg = [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
          [0.0, 0.0034661112, 0.0, 0.0, 0.0004179433, 0.0],
          [0.6585807425, 0.0003363161, 0.8733738848, 0.2962771222, 8.72605e-05, 0.1860096977],
          [0.0, 0.0003499197, 0.0, 0.0, -8.99441e-05, 0.0]]
    dg = np.array(dg)
    model = AtomicGaussianDensity(points, num_s=0, num_p=3, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1s1p_basis_origin():
    r"""Test evaluation of single s-type and p-type Gaussian at origin."""
    # test (un)normalized s-type + p-type gaussian at origin against sympy
    coeffs = np.array([0.5, -2.0])
    expons = np.array([0.1, 0.4])
    # un-normalized at r=0.
    g, dg = np.array([0.5]), np.array([[1., 0., 0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g, dg = np.array([0.0028395217]), np.array([[0.0056790434, 0., 0.04259282582, 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1s1p_basis_one_point():
    r"""Test evaluation of single s-type and p-type Gaussian at single point."""
    # test (un)normalized s-type & p-type gaussian on one point against sympy
    coeffs = np.array([0.5, 2.0])
    expons = np.array([0.1, 0.4])
    # un-normalized at r=-3.
    g, = np.array([0.6951118339])
    dg = np.array([[0.4065696597, 0.2459135020, -1.8295634688, -4.4264430364]])
    model = AtomicGaussianDensity(np.array([-3.]), num_s=1, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=-3.
    g = np.array([0.0071130914])
    dg = np.array([[0.0023089267, 0.0029793140, 0.0069267802, -0.0163862272]])
    model = AtomicGaussianDensity(np.array([-3.]), num_s=1, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_2s1p_basis_origin():
    r"""Test evaluation of two s-type and one p-type Gaussian at origin."""
    # test (un)normalized s-type & p-type gaussian on one point against sympy
    coeffs = np.array([1.2, -4.5, -0.75])
    expons = np.array([0.8, 1.25, 2.19])
    # un-normalized at r=0.
    g, dg = np.array([-3.3]), np.array([[1., 1., 0., 0., 0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=2, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g = np.array([-0.9752103479])
    dg = np.array([[0.1285020841, 0.2509806330, 0., 0.2891296893, -1.3552954187, 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=2, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_2s1p_basis_multiple_point():
    r"""Test evaluation of two s-type and one p-type Gaussian at multiple points."""
    # test (un)normalized s-type & p-type gaussian on one point against sympy
    coeffs = np.array([1.2, -4.5, -0.75])
    expons = np.array([0.8, 1.25, 2.19])
    # un-normalized at r=[0.89, -2.56]
    g = np.array([-1.1399500497, 0.0050932753])
    dg = np.array([[5.30635465e-01,  3.71530247e-01,  1.39768663e-01,
                    -5.04379622e-01,  1.32430099e+00,  8.30330683e-02],
                   [5.28501406e-03,  2.76859611e-04,  3.83105827e-06,
                    -4.15630417e-02,  8.16492216e-03,  1.88304176e-05]])
    model = AtomicGaussianDensity(np.array([0.89, -2.56]), num_s=2, num_p=1, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=[0.89, -2.56]
    g = np.array([-0.4268626593, 0.000499832])
    dg = np.array([[6.81877632e-02,  9.32468966e-02,  1.18769254e-01,
                    8.86086345e-02, -1.71159341e-01, -3.11281606e-02],
                   [6.79135321e-04,  6.94864004e-05,  3.25546459e-06,
                    -3.81288302e-03,  1.67401077e-03,  1.32140467e-05]])
    model = AtomicGaussianDensity(np.array([0.89, -2.56]), num_s=2, num_p=1, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1s2p_basis_origin():
    r"""Test evaluation of one s-type and two p-type Gaussian at origin."""
    # test (un)normalized s-type & p-type gaussian on one point against sympy
    coeffs = np.array([1.2, -4.5, -0.75])
    expons = np.array([0.8, 1.25, 2.19])
    # un-normalized at r=0.
    g, dg = np.array([1.2]), np.array([[1., 0., 0., 0., 0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=2, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=0.
    g = np.array([0.1542025009])
    dg = np.array([[0.1285020841, 0., 0., 0.2891296893, 0., 0.]])
    model = AtomicGaussianDensity(np.array([0.]), num_s=1, num_p=2, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_1s2p_basis_multiple_point():
    r"""Test evaluation of one s-type and two p-type Gaussian at multiple points."""
    # test (un)normalized s-type & p-type gaussian on one point against sympy
    coeffs = np.array([1.2, -4.5, -0.75])
    expons = np.array([0.8, 1.25, 2.19])
    # un-normalized at r=[-0.201, 1.701]
    g = np.array([0.9612490198, -0.2351426929])
    dg = np.array([[0.96819593, 0.03841136, 0.03697997, -0.04693930, 0.00698336, 0.00112052],
                   [0.09879376, 0.07774519, 0.00512249, -0.34301997, 1.01226602, 0.01111605]])
    model = AtomicGaussianDensity(np.array([-0.201, 1.701]), num_s=1, num_p=2, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized at r=[-0.201, 1.701]
    g = np.array([0.0895783695, -0.0612024131])
    dg = np.array([[0.1244152, 0.00803376, 0.03142395,  0.27390239, -0.07084324, -0.0259519],
                   [0.0126952, 0.01626045, 0.00435286, -0.01551457,  0.06537195,  0.00571917]])
    model = AtomicGaussianDensity(np.array([-0.201, 1.701]), num_s=1, num_p=2, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_2s2p_basis_multiple_point():
    r"""Test evaluation of two s-type and two p-type Gaussian at origin."""
    # test multiple (un)normalized s-type & p-type gaussian on one point against sympy
    coeffs = np.array([0.53, -6.1, -2.0, 4.3])
    expons = np.array([0.11, 3.4, 1.5, 0.78])
    points = np.array([0., -0.71, 1.68, -2.03, 3.12, 4.56])
    # un-normalized
    g = np.array([-5.57, 0.3920707885, 1.6490462639, 1.0318279643, 0.2027417225, 0.0538246312])
    dg = np.array([[1.00000000e+00, 1.00000000e+00, 0.00000000e+00,  0.00000000e+00,
                    -0.00000000e+00, 0.00000000e+00, 0.00000000e+00, -0.00000000e+00],
                   [9.46058379e-01, 1.80154583e-01, 2.36660035e-01,  3.40214622e-01,
                    -2.52761255e-01, 5.53977145e-01, 2.38600648e-01, -7.37459422e-01],
                   [7.33106716e-01, 6.79893148e-05, 4.09250713e-02,  3.12270833e-01,
                    -1.09663381e+00, 1.17054756e-03, 2.31013843e-01, -3.78981876e+00],
                   [6.35528082e-01, 8.22382682e-07, 8.52051568e-03,  1.65591658e-01,
                    -1.38804227e+00, 2.06726365e-05, 7.02243861e-02, -2.93426265e+00],
                   [3.42739704e-01, 4.22836923e-15, 4.43521708e-06,  4.90663947e-03,
                    -1.76827365e+00, 2.51079889e-13, 8.63483543e-05, -2.05381723e-01],
                   [1.01540657e-01, 1.97762748e-31, 5.91712954e-13,  1.87976695e-06,
                    -1.11903978e+00, 2.50844168e-29, 2.46076850e-11, -1.68074625e-04]])
    dg = np.array(dg)
    model = AtomicGaussianDensity(points, num_s=2, num_p=2, normalize=False)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized
    g = np.array([-6.8644186384, -1.2960445644, 0.0614559599,
                  0.0423855159, 0.0025445224, 0.0003531182])
    dg = np.array([[6.55185411e-03,  1.12588379e+00,  0.00000000e+00,  0.00000000e+00,
                    4.73520366e-02, -3.02995197e+00,  0.00000000e+00,  0.00000000e+00],
                   [6.19843649e-03,  2.02833125e-01,  7.80794966e-02,  2.18863537e-02,
                    4.31417361e-02,  7.78541543e-02, -1.81545240e-01,  2.54197333e-01],
                   [4.80320825e-03,  7.65480675e-05,  1.35021064e-02,  2.00887013e-02,
                    2.75291113e-02,  1.11189616e-03,  3.12096688e-02,  3.30606041e-02],
                   [4.16388728e-03,  9.25907331e-07,  2.81111077e-03,  1.06526803e-02,
                    2.09992985e-02,  2.07832063e-05,  1.37982435e-02, -4.19483238e-02],
                   [2.24558054e-03,  4.76065238e-15,  1.46327840e-06,  3.15649123e-04,
                    4.64395201e-03,  2.69875021e-13,  2.36106798e-05, -8.86212334e-03],
                   [6.65279571e-04,  2.22657872e-31,  1.95219482e-13,  1.20927325e-07,
                    -2.52362846e-03,  2.76429266e-29,  7.46790004e-12, -9.14578547e-06]])
    dg = np.array(dg)
    model = AtomicGaussianDensity(points, num_s=2, num_p=2, normalize=True)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_molecular_gaussian_density_1d_1center_1s():
    r"""Test evaluation of one s-type Molecular Gaussian at origin."""
    # points in 1D space
    points = np.array([0.0, 1.0, 2.0, 3.0, 4.0])
    # gaussian parameters
    coeffs = np.array([1.5])
    expons = np.array([0.5])
    coords = np.array([[2.5]])
    # un-normalized 1s basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[1, 0]]), normalize=False)
    # check basis & density
    assert_equal(model.nbasis, 1)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_raises(ValueError, model.assign_basis_to_center, 1)
    g = np.array([0.0659054004, 0.486978701, 1.3237453539, 1.3237453539, 0.486978701])
    dg = np.array([[0.0439369336, -0.4119087527], [0.3246524674, -1.0957020773],
                   [0.8824969026, -0.3309363385], [0.8824969026, -0.3309363385],
                   [0.3246524674, -1.0957020773]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized 1s basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[1, 0]]), normalize=True)
    # check basis & density
    assert_equal(model.nbasis, 1)
    assert_equal(model.assign_basis_to_center(0), 0)
    g = np.array([0.0041845735, 0.0309200484, 0.0840494056, 0.0840494056, 0.0309200484])
    dg = np.array([[0.0027897157, -0.0135998639], [0.0206133656, 0.0231900363],
                   [0.0560329370,  0.2311358653], [0.0560329370, 0.2311358653],
                   [0.0206133656,  0.0231900363]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_molecular_gaussian_density_1d_1center_2s():
    r"""Test evaluation of two s-type Molecular Gaussian at origin."""
    # points in 1D space
    points = np.array([-3.09, -1.53, 0.0, 0.79, 2.91, 4.09])
    # gaussian parameters
    coeffs = np.array([6.03, 2.56])
    expons = np.array([2.45, 0.36])
    coords = np.array([[1.45]])
    # un-normalized 1s basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[2, 0]]), normalize=False)
    # check basis & density
    assert_equal(model.nbasis, 2)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    g = np.array(
        [0.0015335519, 0.1046706295, 1.2358743034, 4.2625443763, 1.2209551847, 0.2082434127]
    )
    dg = np.array([[1.17169574e-22, 5.99043705e-04, -1.45627659e-20, -3.16089580e-02],
                   [3.55683438e-10, 4.08869638e-02, -1.90464255e-08, -9.29517039e-01],
                   [5.79288407e-03, 4.69118442e-01, -7.34426186e-02, -2.52498310e+00],
                   [3.43963408e-01, 8.54861339e-01, -9.03477676e-01, -9.53286654e-01],
                   [5.39425921e-03, 4.64229610e-01, -6.93353697e-02, -2.53325270e+00],
                   [3.83880356e-08, 8.13449927e-02, -1.61332199e-06, -1.45137168e+00]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized 1s basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[2, 0]]), normalize=True)
    # check basis & density
    assert_equal(model.nbasis, 2)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    g = np.array([5.94877e-05, 0.0040602608, 0.070642293, 1.5133048321, 0.0685013983, 0.0080780828])
    dg = np.array([[8.06936140e-23, 2.32373955e-05, -9.73133606e-21, -9.78271797e-04],
                   [2.44955932e-10, 1.58603879e-03, -1.22127618e-08, -1.91389796e-02],
                   [3.98950629e-03, 1.81974883e-02, -3.58506541e-02,  9.61603808e-02],
                   [2.36884454e-01, 3.31607709e-02,  2.52321914e-01,  3.16736121e-01],
                   [3.71497700e-03, 1.80078465e-02, -3.40355466e-02,  9.38167504e-02],
                   [2.64374891e-08, 3.15543884e-03, -1.01347721e-06, -2.26418808e-02]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_molecular_gaussian_density_1d_1center_1s2p():
    r"""Test evaluation of one s-type and two-p-type Molecular Gaussian at origin."""
    # points in 1D space
    points = np.array([-2.22, -0.76, 0.0, 1.98, 3.25, 5.50])
    # gaussian parameters
    coeffs = np.array([1.45, 0.0, 1.45])
    expons = np.array([0.75, 1.0, 0.75])
    coords = np.array([[1.00]])
    # un-normalized 1s + 2p basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[1, 2]]), normalize=False)
    # check basis & density
    assert_equal(model.nbasis, 3)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    assert_equal(model.assign_basis_to_center(2), 0)
    g = np.array([0.0069161303, 0.5820289984, 1.369863003, 1.3832172004, 0.1972685684, 7.8141e-06])
    dg = np.array([[0.0004195617, 0.0003256663, 0.0043501834, -0.0063077659, 0.0, -0.0654014398],
                   [0.0979596128, 0.1398797181, 0.3034396965, -0.4399875599, 0.0, -1.3629054655],
                   [0.4723665527, 0.3678794412, 0.4723665527, -0.6849315015, 0.0, -0.6849315015],
                   [0.4866062522, 0.3675832650, 0.4673366446, -0.6776381347, 0.0, -0.6508036645],
                   [0.0224407899, 0.0320441844, 0.1136064987, -0.1647294231, 0.0, -0.8339427044],
                   [2.536e-07, 3.25e-08, 5.1354e-06, -7.4464e-06, 0.0, -0.0001507893]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized 1s + 2p basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[1, 2]]), normalize=True)
    # check basis & density
    assert_equal(model.nbasis, 3)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    assert_equal(model.assign_basis_to_center(2), 0)
    g = np.array([0.0004388483, 0.0422296913, 0.119841017, 0.1218240891, 0.01340299, 4.772e-07])
    dg = np.array([[4.89399e-05, 3.89903e-05, 0.0002537141, -0.0005938453, 0.0, -0.0025880989],
                   [0.0114265243, 0.016747064, 0.0176974008, -0.0181855419, 0.0, 0.0060492075],
                   [0.0550993182, 0.0440442734, 0.0275496591, 0.0798940113, 0.0, 0.0932096799],
                   [0.0567603116, 0.0440088138, 0.0272563016, 0.0855616289, 0.0, 0.0937822106],
                   [0.0026176117, 0.0038364819, 0.0066258296, -0.011623832, 0.0, -0.0166128875],
                   [2.96e-08, 3.9e-09, 2.995e-07, -7.828e-07, 0.0, -7.3468e-06]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_molecular_gaussian_density_2d_1center_2s1p():
    r"""Test evaluation of two s-type and single p-type Molecular Gaussian."""
    # points in 2D space
    points = np.array([[0.0, 0.0], [0.25, -0.25], [-0.5, -0.5], [0.25, 0.5]])
    # gaussian parameters
    coeffs = np.array([1.45, 0.0, 1.45])
    expons = np.array([0.75, 1.0, 0.75])
    coords = np.array([[0.5, 0.5]])
    # un-normalized 2s + 1p basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[2, 1]]), normalize=False)
    g = np.array([1.4948541813, 1.4745035726, 0.9706161966, 1.4700746448])
    dg = np.array([[0.6872892788, 0.6065306597, 0.3436446394, -0.4982847271, 0.0, -0.2491423636],
                   [0.6257840096, 0.5352614285, 0.391115006, -0.5671167587, 0.0, -0.3544479742],
                   [0.2231301601, 0.1353352832, 0.4462603203, -0.6470774644, 0.0, -1.2941549289],
                   [0.954206666, 0.9394130628, 0.0596379166, -0.0864749791, 0.0, -0.0054046862]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized 2s + 1p basis function
    model = MolecularGaussianDensity(points, coords, basis=np.array([[2, 1]]), normalize=True)
    g = np.array([0.1453063757, 0.1389181087, 0.0754785174, 0.1664337873])
    dg = np.array([[0.0801690349, 0.1089250957, 0.0200422587, 0.1743676509, 0.0, 0.0823402796],
                   [0.0729947369, 0.0961260595, 0.0228108553, 0.1455332567, 0.0, 0.0895801296],
                   [0.026027075, 0.024304474, 0.026027075, 0.0, 0.0, 0.0503190116],
                   [0.1113036822, 0.1687064884, 0.0034782401, 0.3126937823, 0.0, 0.0164962782]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_molecular_gaussian_density_2d_2center_1s1p_2p():
    r"""Test evaluation of Molecular Gaussian of multiple centers."""
    # points in 2D space
    points = np.array([[0.0, 0.0], [0.25, -0.25], [-0.5, -0.5], [0.25, 0.5]])
    # gaussian parameters
    coeffs = np.array([1.45, 0.45, 2.91, 0.89])
    expons = np.array([0.75, 0.65, 1.23, 0.00])
    basis = np.array([[1, 1], [0, 2]])
    coords = np.array([[0.0, 0.0], [0.5, 0.5]])
    # un-normalized (1s + 1p) & (2p) basis functions
    model = MolecularGaussianDensity(points, coords, basis=basis, normalize=False)
    # check basis & density
    assert_equal(model.nbasis, 4)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    assert_equal(model.assign_basis_to_center(2), 1)
    assert_equal(model.assign_basis_to_center(3), 1)
    g = np.array([2.6816325027, 2.7715087598, 3.4363695234, 1.4858614587])
    dg = [[1.0, 0.0, 0.2703204, 0.5, 0.0, 0.0, -0.3933163, -0.2225],
          [0.9105104, 0.1152454, 0.2897451, 0.625, -0.16503, -0.0064826, -0.5269739, -0.3476563],
          [0.6872893, 0.3612637, 0.1708699, 2.0, -0.4982847, -0.0812843, -0.9944628, -3.56],
          [0.7910651, 0.2550551, 0.0578753, 0.0625, -0.3584514, -0.0358671, -0.0105261, -0.0034766]]
    dg = np.array(dg)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=7)
    assert_raises(ValueError, model.evaluate, np.array([[1.]]), np.array([1.]))
    assert_raises(ValueError, model.evaluate, np.array([1., 2.]), np.array([1.]))
    # normalized (1s + 1p) & (2p) basis functions
    model = MolecularGaussianDensity(points, coords, basis=basis, normalize=True)
    # check basis & density
    assert_equal(model.nbasis, 4)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    assert_equal(model.assign_basis_to_center(2), 1)
    assert_equal(model.assign_basis_to_center(3), 1)
    g = np.array([0.3271580029, 0.3254922487, 0.2227611062, 0.1723104634])
    dg = np.array([[0.1166453, 0.0, 0.0543032, 0.0, 0.3382712, 0.0, 0.2421725, 0.0],
                   [0.1062067, 0.0046999, 0.0582053, 0.0, 0.2887495, 0.0078701, 0.2384023, 0.0],
                   [0.080169, 0.014733, 0.0343251, 0.0, 0.1743677, 0.0221845, 0.0032483, 0.0],
                   [0.092274, 0.0104016, 0.0116263, 0.0, 0.2257829, 0.0165401, 0.0666506, 0.0]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=7)


def test_molecular_gaussian_density_3d_2center_1p_1p():
    r"""Test evaluation of Molecular Gaussian of multiple centers with p-type Gaussians."""
    # points in 3D space
    points = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, -1.0], [0.0, -1.0, 0.0]])
    # gaussian parameters
    coeffs = np.array([0.6, 1.6])
    expons = np.array([2.5, 0.7])
    coords = np.array([[0.0, 0.0, 0.0], [0.0, 0.5, -0.5]])
    basis = np.array([[0, 1], [0, 1]])
    # un-normalized (1p) & (1p) basis functions
    model = MolecularGaussianDensity(points, coords, basis=basis, normalize=False)
    # check basis & density
    assert_equal(model.nbasis, 2)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 1)
    g = np.array([0.56375047, 0.84793613, 0.74434677])
    dg = np.array([[0.,  0.35234404,  0., -0.28187524],
                   [0.01347589,  0.52490662, -0.01617107, -1.2597759],
                   [0.082085,  0.43443486, -0.049251, -1.73773943]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # un-normalized (1p) & (1p) basis functions
    model = MolecularGaussianDensity(points, coords, basis=basis, normalize=True)
    # check basis & density
    assert_equal(model.nbasis, 2)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 1)
    g = np.array([0.02767043, 0.05078846, 0.09238776])
    dg = np.array([[0.00000000e+00,  1.72940204e-02,  0.00000000e+00, 8.49877575e-02],
                   [1.59437891e-02,  2.57638692e-02, -9.56627343e-03, 8.53888235e-02],
                   [9.71175569e-02,  2.13232647e-02, -1.38777878e-17, 3.65541680e-02]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_molecular_gaussian_density_3d_3center_2s1p_1p_2s():
    r"""Test evaluation of Molecular Gaussian with three centers."""
    # points in 3D space
    points = np.array([[0.5, 0.5, 0.5], [1.0, 0.0, 0.0], [0.0, 1.5, 0.5], [0.0, 0.0, -0.5]])
    # gaussian parameters
    coeffs = np.array([3.2, 1.7, 0.9, 4.1, 2.2, 7.1])
    expons = np.array([0.8, 2.5, 1.9, 1.1, 7.5, 3.1])
    coords = np.array([[0.0, 0.0, 0.0], [1.0, 0.5, -0.5], [0.0, 1.0, 0.0]])
    basis = np.array([[2, 1], [0, 1], [2, 0]])
    # un-normalized (2s1p) & (1p) & (1s) basis functions
    model = MolecularGaussianDensity(points, coords, basis=basis, normalize=False)
    # check basis
    assert_equal(model.nbasis, 6)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    assert_equal(model.assign_basis_to_center(2), 0)
    assert_equal(model.assign_basis_to_center(3), 1)
    assert_equal(model.assign_basis_to_center(4), 2)
    assert_equal(model.assign_basis_to_center(5), 2)
    # compute density value & its derivatives
    g = np.array([4.177243712, 2.9091656819, 2.4681836067, 5.1131523036])
    dg = np.array([[5.48811636e-01,  1.53354967e-01,  1.80381347e-01,
                    3.16049495e-01,  3.60656314e-03,  9.77834441e-02,
                    -1.31714793e+00, -1.95527583e-01, -1.21757409e-01,
                    -1.61975366e+00, -5.95082917e-03, -5.20696840e-01],
                   [4.49328964e-01,  8.20849986e-02,  1.49568619e-01,
                    2.88474905e-01,  3.05902321e-07,  2.02943064e-03,
                    -1.43785269e+00, -1.39544498e-01, -1.34611757e-01,
                    -5.91373556e-01, -1.34597021e-06, -2.88179150e-02],
                   [1.35335283e-01,  1.93045414e-03,  2.16292380e-02,
                    1.10649502e-01,  2.35177459e-02,  2.12247974e-01,
                    -1.08268227e+00, -8.20443008e-03, -4.86657855e-02,
                    -1.36098888e+00, -2.58695204e-02, -7.53480307e-01],
                   [8.18730753e-01,  5.35261429e-01,  1.55471264e-01,
                    3.16049495e-01,  8.48182352e-05,  2.07543379e-02,
                    -6.54984602e-01, -2.27486107e-01, -3.49810344e-02,
                    -1.61975366e+00, -2.33250147e-04, -1.84194749e-01]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)
    # normalized (2s1p) & (1p) & (1s) basis functions
    model = MolecularGaussianDensity(points, coords, basis=basis, normalize=True)
    # check basis
    assert_equal(model.nbasis, 6)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 0)
    assert_equal(model.assign_basis_to_center(2), 0)
    assert_equal(model.assign_basis_to_center(3), 1)
    assert_equal(model.assign_basis_to_center(4), 2)
    assert_equal(model.assign_basis_to_center(5), 2)
    g = np.array([1.4141296252, 0.5578528034, 1.8064862846, 1.4079886645])
    dg = np.array([[7.05234390e-02,  1.08863690e-01,  1.07463149e-01,
                    4.80198440e-02,  1.33033380e-02,  9.58480014e-02,
                    2.53884381e-01, -2.77602409e-02,  5.47213665e-02,
                    2.01355937e-01, -1.60970390e-02, -1.81106344e-01],
                   [5.77397084e-02,  5.82705342e-02,  8.91063016e-02,
                    4.38302234e-02,  1.12836565e-06,  1.98926180e-03,
                    1.61671183e-01, -3.96239632e-02,  2.53249489e-02,
                    3.18566033e-01, -4.46832798e-06, -2.14134408e-02],
                   [1.73908660e-02,  1.37039161e-03,  1.28857338e-02,
                    1.68118346e-02,  8.67486608e-02,  2.08046917e-01,
                    -3.47817319e-02, -4.42636491e-03, -1.37334794e-02,
                    -5.01298341e-02, -5.72541161e-02, -2.38247276e-02],
                   [1.05208608e-01,  3.79971613e-01,  9.26228338e-02,
                    4.80198440e-02,  3.12864522e-04,  2.03435441e-02,
                    5.47084762e-01,  2.26083110e-01,  8.88447971e-02,
                    2.01355937e-01, -7.22717046e-04, -1.10659036e-01]])
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=False), decimal=8)
    assert_almost_equal(g, model.evaluate(coeffs, expons, deriv=True)[0], decimal=8)
    assert_almost_equal(dg, model.evaluate(coeffs, expons, deriv=True)[1], decimal=8)


def test_gaussian_model_s_integrate_uniform():
    r"""Test integration of one s-type Gaussian with uniform grid against analytic value."""
    grid = UniformRadialGrid(100, 0.0, 15.0)
    spherical = 4.0 * np.pi * grid.points**2.0
    # un-normalized s-type Gaussian at origin
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=0, normalize=False)
    # check center
    assert_almost_equal(model.coord, np.zeros(1), decimal=8)
    # check integration
    value = model.evaluate(np.ones(1), np.ones(1), deriv=False)
    assert_almost_equal(grid.integrate(value * spherical), np.pi**1.5, decimal=6)
    value = model.evaluate(np.array([1.23]), np.array([0.5]), deriv=False)
    assert_almost_equal(grid.integrate(value * spherical), 1.23 * (np.pi / 0.5)**1.5, decimal=6)
    value += model.evaluate(np.array([0.91]), np.array([0.1]), deriv=False)
    value = grid.integrate(value * spherical)
    assert_almost_equal(value, 1.23 * (np.pi / 0.5)**1.5 + 0.91 * (np.pi / 0.1)**1.5, decimal=6)


def test_gaussian_model_p_integrate_uniform():
    r"""Test integration of one p-type Gaussian with uniform grid against analytic value."""
    grid = UniformRadialGrid(100, 0.0, 15.0)
    spherical = 4.0 * np.pi * grid.points**2.0
    # un-normalized p-type Gaussian at origin
    model = AtomicGaussianDensity(grid.points, num_s=0, num_p=1, normalize=False)
    # check center
    assert_almost_equal(model.coord, np.zeros(1), decimal=8)
    # check integration
    value = model.evaluate(np.array([1.0]), np.array([2.5]), deriv=False)
    assert_almost_equal(grid.integrate(value * spherical), 1.5 * np.pi**1.5 / 2.5**2.5, decimal=6)
    value = model.evaluate(np.array([1.2]), np.array([1.0]), deriv=False)
    assert_almost_equal(grid.integrate(value * spherical), 1.2 * 1.5 * np.pi**1.5, decimal=6)
    value = model.evaluate(np.array([2.5]), np.array([1.1]), deriv=False)
    assert_almost_equal(
        grid.integrate(value * spherical), 2.5 * 1.5 * np.pi**1.5 / 1.1**2.5, decimal=6
    )
    # normalized p-type Gaussian at origin
    model = AtomicGaussianDensity(grid.points, num_s=0, num_p=1, normalize=True)
    # check center
    assert_almost_equal(model.coord, np.zeros(1), decimal=8)
    # check integration
    value = model.evaluate(np.array([2.]), np.array([2.75]), deriv=False)
    assert_almost_equal(grid.integrate(value * spherical), 2.0 * 1.0, decimal=6)


def test_gaussian_model_sp_integrate_uniform():
    r"""Test integration of one s/p-type Gaussian with uniform grid against analytic value."""
    grid = UniformRadialGrid(100, 0.0, 15.0)
    spherical = 4.0 * np.pi * grid.points**2.0
    # un-normalized s-type + p-type Gaussian at origin
    model = AtomicGaussianDensity(grid.points, num_s=1, num_p=1, normalize=False)
    # check center
    assert_almost_equal(model.coord, np.zeros(1), decimal=8)
    # check integration
    value = model.evaluate(np.array([1., 1.]), np.array([4.01, 1.25]), deriv=False)
    value = grid.integrate(value * spherical)
    assert_almost_equal(value, (np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5), decimal=6)
    # normalized s-type + p-type Gaussian at origin
    model = AtomicGaussianDensity(grid.points, num_s=2, num_p=3, normalize=True)
    # check center
    assert_almost_equal(model.coord, np.zeros(1), decimal=8)
    # check integration
    coeffs = np.array([1.05, 3.62, 0.56, 2.01, 3.56])
    expons = np.array([0.50, 1.85, 0.16, 1.36, 2.08])
    value = model.evaluate(coeffs, expons, deriv=False)
    assert_almost_equal(grid.integrate(value * spherical), np.sum(coeffs), decimal=6)


def test_gaussian_model_s_integrate_cubic():
    r"""Test integration of one s-type Molecular Gaussian with cubic grid."""
    axes = np.array([[0.3, 0.0, 0.0], [0.0, 0.3, 0.0], [0.0, 0.0, 0.3]])
    grid = CubicGrid(np.array([-5.0, -5.0, -5.0]), axes, (65, 65, 65))
    coord = None
    # un-normalized s-type + p-type Gaussian
    model = AtomicGaussianDensity(grid.points, center=coord, num_s=1)
    # check center
    assert_almost_equal(model.coord, np.array([0., 0., 0.]), decimal=8)
    # check integration
    value = model.evaluate(np.array([1.]), np.array([4.01]), deriv=False)
    value = grid.integrate(value)
    assert_almost_equal(value, (np.pi / 4.01) ** 1.5, decimal=6)
    # check assertion error on types of coefficients
    assert_raises(ValueError, model.evaluate, np.array([[1.]]), np.array([1.]))
    assert_raises(ValueError, model.evaluate, np.array([1., 2.]), np.array([1.]))
    assert_raises(ValueError, model.evaluate, np.array([1., 2.]), np.array([1., 2.]))


def test_gaussian_model_sp_integrate_cubic():
    r"""Test integration of one s-type and one p-type Molecular Gaussian with cubic grid."""
    axes = np.array([[0.3, 0.0, 0.0], [0.0, 0.3, 0.0], [0.0, 0.0, 0.3]])
    grid = CubicGrid(np.zeros(3), axes, (65, 65, 65))
    coord = np.array([10., 10., 10.])
    # un-normalized s-type + p-type Gaussian
    model = AtomicGaussianDensity(grid.points, center=coord, num_s=1, num_p=1)
    # check center
    assert_almost_equal(model.coord, coord, decimal=8)
    # check integration
    value = model.evaluate(np.array([1., 1.]), np.array([4.01, 1.25]), deriv=False)
    value = grid.integrate(value)
    assert_almost_equal(value, (np.pi / 4.01)**1.5 + 1.5 * (np.pi**1.5 / 1.25**2.5), decimal=6)
    # normalized s-type + p-type Gaussian
    model = AtomicGaussianDensity(grid.points, center=coord, num_s=2, num_p=3, normalize=True)
    # check center
    assert_almost_equal(model.coord, coord, decimal=8)
    # check integration
    coeffs = np.array([1.05, 3.62, 0.56, 2.01, 3.56])
    expons = np.array([0.50, 1.85, 0.16, 1.36, 2.08])
    value = model.evaluate(coeffs, expons, deriv=False)
    assert_almost_equal(grid.integrate(value), np.sum(coeffs), decimal=6)


def test_molecular_gaussian_model_sp_integrate_cubic():
    r"""Test integration of one s-type and one p-type Molecular Gaussian with cubic grid."""
    axes = np.array([[0.3, 0.0, 0.0], [0.0, 0.3, 0.0], [0.0, 0.0, 0.3]])
    grid = CubicGrid(np.zeros(3), axes, (75, 75, 75))
    coord = np.array([[10., 10., 9.5], [10., 10., 10.5]])
    # un-normalized s-type & p-type Gaussian
    basis = np.array([[0, 1], [1, 0]])
    model = MolecularGaussianDensity(grid.points, coord, basis=basis, normalize=False)
    # check basis
    assert_equal(model.nbasis, 2)
    assert_equal(model.assign_basis_to_center(0), 0)
    assert_equal(model.assign_basis_to_center(1), 1)
    # check integration
    value = model.evaluate(np.array([1., 1.]), np.array([1.25, 4.01]), deriv=False)
    value = grid.integrate(value)
    assert_almost_equal(value, 1.5 * (np.pi**1.5 / 1.25**2.5) + (np.pi / 4.01)**1.5, decimal=6)
    # normalized s-type & p-type Gaussian
    basis = np.array([[2, 0], [0, 3]])
    model = MolecularGaussianDensity(grid.points, coords=coord, basis=basis, normalize=True)
    # check integration
    coeffs = np.array([1.05, 3.62, 0.56, 2.01, 3.56])
    expons = np.array([0.50, 1.85, 0.16, 1.36, 2.08])
    value = model.evaluate(coeffs, expons, deriv=False)
    assert_almost_equal(grid.integrate(value), np.sum(coeffs), decimal=6)
    # normalized (s-type + p-type) & p-type Gaussian
    basis = np.array([[2, 2], [0, 1]])
    model = MolecularGaussianDensity(grid.points, coords=coord, basis=basis, normalize=True)
    # check integration
    value = model.evaluate(coeffs, expons, deriv=False)
    assert_almost_equal(grid.integrate(value), np.sum(coeffs), decimal=6)



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Test bfit._slater module."""

import numpy as np

from bfit._slater import load_slater_wfn


def test_parsing_slater_density_be():
    r"""Test correct parsing of beryllium Slater file."""
    # Load the Be file
    be = load_slater_wfn("be")

    assert be['configuration'] == '1S(2)2S(2)'
    assert be['energy'] == -14.573023167
    assert be['kinetic_energy'] == 14.573023130

    # Check basis of S orbitals
    assert be['orbitals'] == ['1S', '2S']
    assert np.all(abs(be['orbitals_cusp'] - np.array([1.0001235, 0.9998774])[:, None]) < 1.e-6)
    assert np.all(abs(be['orbitals_energy'] - np.array([-4.7326699, -0.3092695])[:, None]) < 1.e-6)
    assert be['orbitals_basis']['S'] == ['1S', '1S', '1S', '1S', '1S', '1S', '2S', '1S']
    assert len(be['orbitals_occupation']) == 2
    assert (be['orbitals_occupation'] == np.array([[2], [2]])).all()
    basis_numbers = np.array([[1], [1], [1], [1], [1], [1], [2], [1]])
    assert (be['basis_numbers']['S'] == basis_numbers).all()

    # Check exponents of S orbitals
    exponents = np.array([12.683501, 8.105927, 5.152556, 3.472467, 2.349757,
                          1.406429, 0.821620, 0.786473])
    assert (abs(be['orbitals_exp']['S'] - exponents.reshape(8, 1)) < 1.e-6).all()

    # Check coefficients of S orbitals
    coeff_1s = np.array([-0.0024917, 0.0314015, 0.0849694, 0.8685562, 0.0315855,
                         -0.0035284, -0.0004149, .0012299])
    assert be['orbitals_coeff']['1S'].shape == (8, 1)
    assert (abs(be['orbitals_coeff']['1S'] - coeff_1s.reshape(8, 1)) < 1.e-6).all()
    coeff_2s = np.array([0.0004442, -0.0030990, -0.0367056, 0.0138910, -0.3598016,
                         -0.2563459, 0.2434108, 1.1150995])
    assert be['orbitals_coeff']['2S'].shape == (8, 1)
    assert (abs(be['orbitals_coeff']['2S'] - coeff_2s.reshape(8, 1)) < 1.e-6).all()


def test_parsing_slater_density_ag():
    r"""Test correct parsing of silver Slater file."""
    # Load the Ag file.
    ag = load_slater_wfn("ag")

    # Check configuration and energy.
    assert ag['configuration'] == 'K(2)L(8)M(18)4S(2)4P(6)5S(1)4D(10)'
    assert ag['energy'] == -5197.698467674
    assert ag["kinetic_energy"] == 5197.698468984
    assert ag["potential_energy"] == -10395.396936658

    # Check orbitals
    assert ag['orbitals'] == ['1S', '2S', '3S', '4S', '5S', '2P', '3P', '4P', '3D', '4D']

    # Check basis
    assert ag['orbitals_basis']['P'] == ['2P', '3P', '3P', '2P', '3P', '3P', '3P',
                                         '3P', '2P', '2P', '2P']
    cusp = np.array([1.0002457, 1.0000318, 1.0004188, 1.0004755, 1.0007044,
                     1.0008130, 1.0008629, 0.9998751, 0.9991182, 1.0009214])[:, None]
    energy = np.array([-913.8355964, -134.8784068, -25.9178242, -4.0014988, -0.2199797,
                       -125.1815809, -21.9454343, -2.6768201, -14.6782003, -0.5374007])[:, None]
    assert (abs(ag['orbitals_cusp'] - cusp) < 1.e-6).all()
    assert (abs(ag['orbitals_energy'] - energy) < 1.e-6).all()

    # Check exponents of D orbitals
    exp_d = np.array([53.296212, 40.214567, 21.872645, 17.024065, 10.708021, 7.859216, 5.770205,
                      3.610289, 2.243262, 1.397570, 0.663294])
    assert (abs(ag['orbitals_exp']['D'] - exp_d.reshape(11, 1)) < 1.e-6).all()

    # Check coefficients of 3D orbital
    coeff_3d = np.array([0.0006646, 0.0037211, -0.0072310, 0.1799224, 0.5205360, 0.3265622,
                         0.0373867, 0.0007434, 0.0001743, -0.0000474, 0.0000083])
    assert (abs(ag['orbitals_coeff']['3D'] - coeff_3d.reshape(11, 1)) < 1.e-6).all()

    # Check coefficients of 4D orbital
    coeff_4d = np.array([-0.0002936, -0.0016839, 0.0092799, -0.0743431, -0.1179494, -0.2809146,
                         0.1653040, 0.4851980, 0.4317110, 0.1737644, 0.0013751])
    assert (abs(ag['orbitals_coeff']['4D'] - coeff_4d.reshape(11, 1)) < 1.e-6).all()

    # Check occupation numbers
    assert len(ag['orbitals_occupation']) == 10
    assert (
        ag['orbitals_occupation'] == np.array([2, 2, 2, 2, 1, 6, 6, 6, 10, 10]).reshape(10, 1)
    ).all()


def test_parsing_slater_density_ne():
    r"""Test correct parsing of neon Slater file."""
    # Load the Ne file
    ne = load_slater_wfn("ne")

    assert ne['configuration'] == "1S(2)2S(2)2P(6)"
    assert ne['orbitals'] == ["1S", "2S", "2P"]
    assert (ne["orbitals_occupation"] == np.array([[2], [2], [6]])).all()
    assert ne['energy'] == -128.547098079
    assert ne["kinetic_energy"] == 128.547098140
    assert ne["potential_energy"] == -257.094196219

    # Check orbital energy and cusp
    assert (abs(
        ne['orbitals_energy'] - np.array([-32.7724425, -1.9303907, -0.8504095])[:, None]) < 1.e-6
    ).all()
    assert (abs(
        ne['orbitals_cusp'] - np.array([1.0000603, 0.9996584, 1.0000509])[:, None]) < 1.e-6
    ).all()

    # Check basis
    assert ne['orbitals_basis']['P'] == ['3P', '2P', '3P', '2P', '2P', '2P', '2P']
    assert ne['orbitals'] == ['1S', '2S', '2P']

    # Check exponents of P orbitals
    exp_p = np.array([25.731219, 10.674843, 8.124569, 4.295590, 2.648660, 1.710436, 1.304155])
    assert (abs(ne['orbitals_exp']['P'] - exp_p.reshape(7, 1)) < 1.e-6).all()

    # Check coefficients of P orbitals
    coeff = np.array([0.0000409, 0.0203038, 0.0340866, 0.2801866, 0.3958489, 0.3203928, 0.0510413])
    assert (abs(ne['orbitals_coeff']['2P'] - coeff.reshape(7, 1)) < 1.e-6).all()


def test_parsing_slater_density_h():
    r"""Test correct parsing of hydrogen Slater file."""
    # Load the H file
    h = load_slater_wfn("h")

    assert h['configuration'] == "1S(1)"
    assert h['energy'] == -0.5
    assert h["kinetic_energy"] == 0.5
    assert h["potential_energy"] == -1.0

    # Check orbital energy and cusp
    assert (abs(h['orbitals_energy'] - np.array([-0.50])[:, None]) < 1.e-6).all()
    assert (abs(h['orbitals_cusp'] - np.array([1.])[:, None]) < 1.e-6).all()

    # Check basis
    assert h['orbitals_basis']['S'] == ['1S']
    assert h['orbitals'] == ['1S']

    # Check exponents of S orbitals
    exp_s = np.array([1.])
    assert (abs(h['orbitals_exp']['S'] - exp_s) < 1.e-6).all()

    # Check coefficients of 1S orbitals.
    coeff = np.array([1.])
    assert (abs(h['orbitals_coeff']['1S'] - coeff) < 1.e-6).all()


def test_parsing_slater_density_k():
    r"""Test correct parsing of potassium Slater file."""
    # Load the K file
    k = load_slater_wfn("k")

    assert k["configuration"] == "K(2)L(8)3S(2)3P(6)4S(1)"
    assert k["energy"] == -599.164786322
    assert k["kinetic_energy"] == 599.164786943
    assert k["potential_energy"] == -1198.329573264

    assert k['orbitals'] == ["1S", "2S", "3S", "4S", "2P", "3P"]
    assert np.all(abs(k['orbitals_cusp'] - np.array([1.0003111, 0.9994803, 1.0005849, 1.0001341,
                                                     1.0007902, 0.9998975])[:, None]) < 1.e-6)
    assert np.all(
        abs(k['orbitals_energy'] - np.array([
            -133.5330493, -14.4899575, -1.7487797, -0.1474751, -11.5192795, -0.9544227])[:, None]
            ) < 1.e-6
    )
    assert k['orbitals_basis']['P'] == ['2P', '3P', '2P', '3P', '2P', '2P', '3P', '2P', '2P', '2P']

    basis_numbers = np.array([[2], [3], [2], [3], [2], [2], [3], [2], [2], [2]])
    assert np.all(np.abs(k['basis_numbers']['P'] - basis_numbers) < 1e-5)

    # Check coefficients of 3P orbital
    coeff_3p = np.array([0.0000354, 0.0011040, -0.0153622, 0.0620133, -0.1765320, -0.3537264,
                         -0.3401560, 1.3735350, 0.1055549, 0.0010773])
    assert (abs(k['orbitals_coeff']['3P'] - coeff_3p.reshape(10, 1)) < 1.e-6).all()


def test_parsing_slater_density_i():
    r"""Test correct parsing of iodine Slater file."""
    i = load_slater_wfn("i")
    assert i["configuration"] == "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(5)"
    assert i["orbitals"] == ["1S", "2S", "3S", "4S", "5S", "2P", "3P", "4P", "5P", "3D", "4D"]
    occupation = np.array([[2], [2], [2], [2], [2], [6], [6], [6], [5], [10], [10]])
    assert (i["orbitals_occupation"] == occupation).all()

    exponents = np.array([[58.400845, 45.117174, 24.132009, 20.588554, 12.624386, 10.217388,
                           8.680013, 4.627159, 3.093797, 1.795536, 0.897975]])
    assert np.all(np.abs(i["orbitals_exp"]["D"] - exponents.reshape((11, 1))) < 1e-4)


def test_parsing_slater_density_xe():
    r"""Test correct parsing of xenon Slater file."""
    # Load the Xe file
    xe = load_slater_wfn("xe")
    assert xe['configuration'] == "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(6)"
    assert xe['orbitals'] == ["1S", "2S", "3S", "4S", "5S", "2P", "3P", "4P", "5P", "3D", "4D"]
    occupation = np.array([[2], [2], [2], [2], [2], [6], [6], [6], [6], [10], [10]])
    assert (xe["orbitals_occupation"] == occupation).all()
    assert xe['energy'] == -7232.138355835
    assert xe["kinetic_energy"] == 7232.138367196
    assert xe["potential_energy"] == -14464.276723031

    # Check coeffs of D orbitals.
    coeffs = np.array([-0.0006386, -0.0030974, 0.0445101, -0.1106186, -0.0924762, -0.4855794,
                       0.1699923, 0.7240230, 0.3718553, 0.0251152, 0.0001040])
    assert (abs(xe['orbitals_coeff']["4D"] - coeffs.reshape((len(coeffs), 1))) < 1e-10).all()


def test_parsing_slater_density_xe_cation():
    r"""Test correct parsing of xenon cation Slater file."""
    # Load the Xe file
    np.testing.assert_raises(ValueError, load_slater_wfn, "xe", anion=True)
    xe = load_slater_wfn("xe", cation=True)
    assert xe['configuration'] == "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(5)"
    assert xe['orbitals'] == ["1S", "2S", "3S", "4S", "5S", "2P", "3P", "4P", "5P", "3D", "4D"]
    occupation = np.array([[2], [2], [2], [2], [2], [6], [6], [6], [5], [10], [10]])
    assert (xe["orbitals_occupation"] == occupation).all()
    assert xe['energy'] == -7231.708943297
    assert xe["kinetic_energy"] == 7231.708943551
    assert xe["potential_energy"] == -14463.417886848

    # Check coefficients of D orbitals.
    coeff = np.array([-0.0004316, -0.0016577, -0.0041398, -0.2183952, 0.0051908, -0.2953384,
                     -0.0095762, 0.6460145, 0.4573096, 0.0431928, -0.0000161])
    assert (abs(xe['orbitals_coeff']["4D"] - coeff.reshape((len(coeff), 1))) < 1e-10).all()


def test_parsing_slater_density_c_cation():
    r"""Test correct parsing of carbon cation Slater file."""
    # Load the Xe file
    c = load_slater_wfn("c", cation=True)
    assert c['configuration'] == "1S(2)2S(2)2P(1)"
    assert c['orbitals'] == ["1S", "2S", "2P"]
    occupation = np.array([[2], [2], [1]])
    assert (c["orbitals_occupation"] == occupation).all()
    assert c['energy'] == -37.292223770
    assert c["kinetic_energy"] == 37.292223758
    assert c["potential_energy"] == -74.584447528

    # Check coefficients of P orbitals.
    coeff = np.array([0.0000857, 0.0092705, -0.0120655, 0.0790120, 0.2559441, 0.6807658, 0.0323882])
    assert (abs(c['orbitals_coeff']["2P"] - coeff.reshape((len(coeff), 1))) < 1e-10).all()


def test_parsing_slater_density_h_anion():
    r"""Test correct parsing of hydrogen anion Slater file."""
    # Load the Hydrogen file
    # There is no cation of hydrogen in data folder.
    np.testing.assert_raises(ValueError, load_slater_wfn, "h", cation=True)
    h = load_slater_wfn("h", anion=True)
    assert h['configuration'] == "1S(2)"
    assert h['orbitals'] == ["1S"]
    occupation = np.array([[2]])
    assert (h["orbitals_occupation"] == occupation).all()
    assert h['energy'] == -0.487929734
    assert h["kinetic_energy"] == 0.487929734
    assert h["potential_energy"] == -0.975859469

    # Check coeffs of 1S orbitals.
    coeff = np.array([0.0005803, 0.0754088, 0.2438040, 0.3476471, 0.3357298, 0.0741188])
    assert (abs(h['orbitals_coeff']["1S"] - coeff.reshape((len(coeff), 1))) < 1e-10).all()

    # Check exps of 1S orbitals.
    exps = np.array([3.461036, 1.704290, 1.047762, 0.626983, 0.392736, 0.304047])
    assert (abs(h['orbitals_exp']["S"] - exps.reshape((len(exps), 1))) < 1e-10).all()


def test_parsing_slater_heavy_atom_cs():
    r"""Test correct parsing of cesium Slater file."""
    np.testing.assert_raises(ValueError, load_slater_wfn, "cs", cation=True)
    np.testing.assert_raises(ValueError, load_slater_wfn, "cs", anion=True)
    # Load the Xe file
    cs = load_slater_wfn("cs")
    assert cs['configuration'] == "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(6)6S(1)"
    assert cs['orbitals'] == ["1S", "2S", "3S", "4S", "5S", "6S", "2P", "3P", "4P", "5P", "3D",
                              "4D"]
    occupation = np.array([[2], [2], [2], [2], [2], [1], [6], [6], [6], [6], [10], [10]])
    assert (cs["orbitals_occupation"] == occupation).all()
    assert cs['energy'] == -7553.933536555
    assert cs["kinetic_energy"] == 7553.933539793
    assert cs["potential_energy"] == -15107.867076348

    # Check coeffs of D orbitals.
    coeffs = np.array([-0.0025615, -0.1930536, -0.2057559, -0.1179374, 0.4341816, 0.6417053,
                       0.1309576])
    assert (abs(cs['orbitals_coeff']["4D"] - coeffs.reshape((len(coeffs), 1))) < 1e-10).all()

    # Check Exponents of D orbitals.
    exps = np.array([32.852137, 18.354403, 14.523221, 10.312620, 7.919345, 5.157647,
                     3.330606])
    assert (abs(cs['orbitals_exp']["D"] - exps.reshape((len(exps), 1))) < 1e-10).all()


def test_parsing_slater_heavy_atom_rn():
    r"""Test correct parsing of radom Slater file."""
    # Load the Xe file
    rn = load_slater_wfn("rn")
    assert rn['configuration'] == "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(6)4F(14)6S(2)5D(10)6P(6)"
    print(rn["orbitals"])
    assert rn['orbitals'] == ["1S", "2S", "3S", "4S", "5S", "6S", "2P", "3P", "4P", "5P", "6P",
                              "3D", "4D", "5D", "4F"]
    occupation = np.array([[2], [2], [2], [2], [2], [2], [6], [6], [6], [6], [6], [10], [10],
                           [10], [14]])
    assert (rn["orbitals_occupation"] == occupation).all()
    assert rn['energy'] == -21866.772070663
    assert rn["kinetic_energy"] == 21866.772036482
    assert rn["potential_energy"] == -43733.544107144

    # Check coeffs of 4F orbitals.
    coeffs = np.array([.0196357, .2541992, .4806186, -.2542278, .5847619, .0099519])
    assert (abs(rn['orbitals_coeff']["4F"] - coeffs.reshape((len(coeffs), 1))) < 1e-10).all()


def test_parsing_slater_heavy_atom_lr():
    r"""Test correct parsing of lawrencium Slater file."""
    # Load the lr file
    lr = load_slater_wfn("lr")
    assert lr['configuration'] == "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(6)4F(14)6S(2)5D(10)6P(6)" \
                                  "7S(2)6D(1)5F(14)"
    assert lr['orbitals'] == ["1S", "2S", "3S", "4S", "5S", "6S", "7S",
                              "2P", "3P", "4P", "5P", "6P",
                              "3D", "4D", "5D", "6D", "4F", "5F"]
    occupation = np.array([[2], [2], [2], [2], [2], [2], [2], [6], [6], [6], [6], [6], [10], [10],
                           [10], [1], [14], [14]])
    assert (lr["orbitals_occupation"] == occupation).all()
    assert lr['energy'] == -33557.950142251
    assert lr["kinetic_energy"] == 33557.949960623
    assert lr["potential_energy"] == -67115.900102873

    # Check coeffs of 3D orbitals.
    coeffs = np.array([0.0017317, 0.4240510, 0.4821228, 0.1753365, -0.0207393, 0.0091584,
                       -0.0020913, 0.0005398, -0.0001604, 0.0000443, -0.0000124])
    assert (abs(lr['orbitals_coeff']["3D"] - coeffs.reshape((len(coeffs), 1))) < 1e-10).all()

    # Check coeffs of 2S orbitals.
    coeffs = np.array([0.0386739, -0.4000069, -0.2570804, 1.2022357, 0.0787866, 0.0002957,
                       0.0002277, 0.0002397, -0.0005650, 0.0000087, 0.0000031, -0.0000024,
                       0.0000008, -0.0000002, 0.0000001])
    assert (abs(lr['orbitals_coeff']["2S"] - coeffs.reshape((len(coeffs), 1))) < 1e-10).all()


def test_parsing_slater_heavy_atom_dy():
    r"""Test correct parsing of dysprosium Slater file."""
    # Load the dy file
    dy = load_slater_wfn("dy")
    assert dy['configuration'] == "K(2)L(8)M(18)4S(2)4P(6)5S(2)4D(10)5P(6)6S(2)5D(0)4F(10)"
    assert dy['orbitals'] == ["1S", "2S", "3S", "4S", "5S", "6S",
                              "2P", "3P", "4P", "5P",
                              "3D", "4D", "4F"]
    occupation = np.array([[2], [2], [2], [2], [2], [2], [6], [6], [6], [6], [10], [10], [10]])
    assert (dy["orbitals_occupation"] == occupation).all()
    assert dy['energy'] == -11641.452434124
    assert dy['kinetic_energy'] == 11641.452478555
    assert dy["potential_energy"] == -23282.904912679

    # Check coeffs of 4D orbitals.
    coeffs = np.array([-0.0016462, -0.2087639, -0.2407385, -0.1008913, 0.4844709, 0.6180159,
                       0.1070867])
    assert (abs(dy['orbitals_coeff']["4D"] - coeffs.reshape((len(coeffs), 1))) < 1e-10).all()



# -*- coding: utf-8 -*-
# BFit is a Python library for fitting a convex sum of Gaussian
# functions to any probability distribution
#
# Copyright (C) 2020- The QC-Devs Community
#
# This file is part of BFit.
#
# BFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# BFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# ---
r"""Package for fitting densities to a linear combination of Gaussians."""

from bfit.density import *
from bfit.fit import *
from bfit.greedy import *
from bfit.grid import *
from bfit.measure import *
from bfit.model import *
